#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging

from skale.skale_base import SkaleBase
import skale.contracts.allocator as contracts
from skale.contracts.contract_manager import ContractManager
from skale.utils.contract_info import ContractInfo
from skale.utils.contract_types import ContractTypes
from skale.utils.helper import get_contracts_info


logger = logging.getLogger(__name__)


CONTRACTS_INFO = [
    ContractInfo('contract_manager', 'ContractManager',
                 ContractManager, ContractTypes.API, False),
    ContractInfo('escrow', 'Escrow', contracts.Escrow,
                 ContractTypes.API, True),
    ContractInfo('allocator', 'Allocator', contracts.Allocator,
                 ContractTypes.API, True)
]


def spawn_skale_allocator_lib(skale):
    return SkaleAllocator(skale._endpoint, skale._abi_filepath, skale.wallet)


class SkaleAllocator(SkaleBase):
    def set_contracts_info(self):
        self.init_contract_manager()
        self._SkaleBase__contracts_info = get_contracts_info(CONTRACTS_INFO)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import os

ENV = os.getenv('ENV', 'prod')
ALLOWED_TS_DIFF = int(os.getenv('ALLOWED_TS_DIFF') or 300)
DISABLE_DRY_RUN = os.getenv('DISABLE_DRY_RUN', 'False') == 'True'
DEFAULT_GAS_LIMIT = int(os.getenv('DEFAULT_GAS_LIMIT') or 10 ** 7)
DEFAULT_GAS_PRICE_WEI = int(os.getenv('DEFAULT_GAS_PRICE_WEI') or 0)
LAST_BLOCK_FILE = os.getenv('LAST_BLOCK_FILE')
DEFAULT_GAS_MULTIPLIER = float(os.getenv('DEFAULT_GAS_MULTIPLIER') or 1.2)
DEFAULT_PRIORITY = int(os.getenv('DEFAULT_PRIORITY') or 5)
NO_SYNC_TS_DIFF = int(os.getenv('NO_SYNC_TS_DIFF') or -1)
TXRECORD_EXPIRATION = int(os.getenv('TXRECORD_EXPIRATION') or 24 * 60 * 60)  # 1 day



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.utils.helper import format_fields
from skale.transactions.result import TxRes
from skale.dataclasses.delegation_status import DelegationStatus


FIELDS = [
    'address', 'validator_id', 'amount', 'delegation_period', 'created',
    'started', 'finished', 'info'
]


class DelegationController(BaseContract):
    """Wrapper for DelegationController.sol functions"""

    @format_fields(FIELDS)
    def get_delegation(self, delegation_id: int) -> dict:
        """Returns delegation structure.

        :returns: Info about delegation request
        :rtype: dict
        """
        return self.__raw_get_delegation(delegation_id)

    def get_delegation_full(self, delegation_id: int) -> dict:
        """Returns delegation structure with ID and status fields.

        :returns: Info about delegation request
        :rtype: dict
        """
        delegation = self.get_delegation(delegation_id)
        delegation['id'] = delegation_id
        delegation['status'] = self._get_delegation_status(delegation_id)
        return delegation

    def __raw_get_delegation(self, delegation_id: int) -> list:
        """Returns raw delegation fields.

        :returns: Info about delegation request
        :rtype: list
        """
        return self.contract.functions.getDelegation(delegation_id).call()

    def _get_delegation_ids_by_validator(self, validator_id: int) -> list:
        delegation_ids_len = self._get_delegation_ids_len_by_validator(
            validator_id)
        return [
            self.contract.functions.delegationsByValidator(
                validator_id, _id).call()
            for _id in range(delegation_ids_len)
        ]

    def _get_delegation_ids_by_holder(self, address: str) -> list:
        delegation_ids_len = self._get_delegation_ids_len_by_holder(address)
        return [
            self.contract.functions.delegationsByHolder(address, _id).call()
            for _id in range(delegation_ids_len)
        ]

    def _get_delegation_ids_len_by_validator(self, validator_id: int) -> list:
        return self.contract.functions.getDelegationsByValidatorLength(validator_id).call()

    def _get_delegation_ids_len_by_holder(self, address: str) -> list:
        return self.contract.functions.getDelegationsByHolderLength(address).call()

    def _get_delegation_state_index(self, delegation_id: int) -> str:
        return self.contract.functions.getState(delegation_id).call()

    def _get_delegation_status(self, delegation_id: int) -> str:
        index = self._get_delegation_state_index(delegation_id)
        return DelegationStatus(index).name

    def get_all_delegations(self, delegation_ids: list) -> list:
        """Returns list of formatted delegations with particular status.

        :param delegation_ids: List of delegation IDs
        :type address: list
        :returns: List of formatted delegations
        :rtype: list
        """
        return [
            self.skale.delegation_controller.get_delegation_full(_id)
            for _id in delegation_ids
        ]

    def get_all_delegations_by_holder(self, address: str) -> list:
        """Returns list of formatted delegations for token holder.

        :param address: Ethereum address
        :type address: str
        :returns: List of formatted delegation requests
        :rtype: list
        """
        delegation_ids = self._get_delegation_ids_by_holder(address)
        return self.get_all_delegations(delegation_ids)

    def get_all_delegations_by_validator(self, validator_id: int) -> list:
        """Returns list of formatted delegations for validator.

        :param validator_id: ID of the validator
        :type address: int
        :returns: List of formatted delegations
        :rtype: list
        """
        validator_id = int(validator_id)
        delegation_ids = self._get_delegation_ids_by_validator(validator_id)
        return self.get_all_delegations(delegation_ids)

    @transaction_method
    def delegate(self, validator_id: int, amount: int, delegation_period: int, info: str) -> TxRes:
        """Creates request to delegate amount of tokens to validator_id.

        :param validator_id: ID of the validator to delegate tokens
        :type validator_id: int
        :param amount: Amount of tokens to delegate
        :type amount: int
        :param delegation_period: Period of delegation
        :type delegation_period: int
        :param info: Delegation request information
        :type info: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.delegate(validator_id, amount, delegation_period, info)

    @transaction_method
    def accept_pending_delegation(self, delegation_id: int) -> TxRes:
        """Accepts a pending delegation by delegation ID.

        :param delegation_id: Delegation ID to accept
        :type delegation_id: int
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.acceptPendingDelegation(delegation_id)

    @transaction_method
    def cancel_pending_delegation(self, delegation_id: int) -> TxRes:
        """Cancel pending delegation request.

        :param delegation_id: ID of the delegation to cancel
        :type delegation_id: int
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.cancelPendingDelegation(delegation_id)

    @transaction_method
    def request_undelegation(self, delegation_id: int) -> TxRes:
        """ This method is  for undelegating request in the end of
            delegation period (3/6/12 months)

        :param delegation_id: ID of the delegation to undelegate
        :type delegation_id: int
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.requestUndelegation(delegation_id)

    def get_delegated_to_validator_now(self, validator_id: int) -> int:
        """Amount of delegated tokens to the validator

        :param validator_id: ID of the validator
        :type validator_id: int
        :returns: Amount of delegated tokens
        :rtype: int
        """
        return self.contract.functions.getAndUpdateDelegatedToValidatorNow(validator_id).call()

    def get_delegated_to_validator(self, validator_id: int, month: int) -> int:
        """Amount of delegated tokens to the validator unil month

        :param validator_id: ID of the validator
        :type validator_id: int
        :param month: last requested month
        :type month: int
        :returns: Amount of delegated tokens
        :rtype: int
        """

        return self.contract.functions.getDelegatedToValidator(validator_id, month).call()

    def get_delegated_amount(self, address: str) -> int:
        """Amount of delegated tokens by token holder

        :param address: Token holder address
        :type address: str
        :returns: Amount of delegated tokens
        :rtype: int
        """
        return self.contract.functions.getAndUpdateDelegatedAmount(address).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import functools
import json
import time
import logging
import urllib

import requests

from skale.utils.exceptions import RPCWalletError
from skale.wallets.sgx_wallet import SgxWallet


logger = logging.getLogger(__name__)


ATTEMPTS = 100
TIMEOUTS = [2 ** p for p in range(ATTEMPTS)]


def rpc_request(func):
    @functools.wraps(func)
    def wrapper(self, route, *args, **kwargs):
        data, error, response = None, None, None
        for i, timeout in enumerate(TIMEOUTS):
            logger.info(f'Sending wallet rpc {route} request. Attempt {i}')
            try:
                response = func(self, route, *args, **kwargs).json()
                data, error = response.get('data'), response.get('error')
            except Exception as err:
                error = 'RPC request failed'
                logger.error(error, exc_info=err)

            if isinstance(error, str) and \
                error.startswith('Dry run failed') or \
                    not error or not self._retry_if_failed:
                break

            logger.info(f'Sleeping {timeout}s ...')
            time.sleep(timeout)

        if error is not None:
            raise RPCWalletError(error)
        logger.info(f'Rpc wallet {route} request returned {data}')
        return data
    return wrapper


class RPCWallet(SgxWallet):
    def __init__(self, url, sgx_endpoint, web3, key_name=None,
                 path_to_cert=None, retry_if_failed=False) -> None:
        super().__init__(sgx_endpoint=sgx_endpoint,
                         web3=web3, key_name=key_name,
                         path_to_cert=path_to_cert)
        self._url = url
        self._retry_if_failed = retry_if_failed

    @classmethod
    def _construct_url(cls, host, url):
        return urllib.parse.urljoin(host, url)

    @rpc_request
    def _post(self, route, data):
        request_url = self._construct_url(self._url, route)
        return requests.post(request_url, json=data)

    @rpc_request
    def _get(self, route, data=None):
        request_url = self._construct_url(self._url, route)
        return requests.get(request_url, data=data)

    @classmethod
    def _compose_tx_data(cls, tx_dict):
        return {
            'transaction_dict': json.dumps(tx_dict)
        }

    def sign_and_send(self, tx_dict):
        data = self._post('/sign-and-send',
                          self._compose_tx_data(tx_dict))
        return data['transaction_hash']



from web3.exceptions import ContractLogicError


class TransactionError(Exception):
    pass


class DryRunFailedError(TransactionError):
    pass


class InsufficientBalanceError(TransactionError):
    pass


class TransactionNotSentError(TransactionError):
    pass


class TransactionNotMinedError(TimeoutError, TransactionError):
    pass


class TransactionFailedError(TransactionError):
    pass


class RevertError(TransactionError, ContractLogicError):
    pass



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2021 SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass


@dataclass
class SchainOptions:
    multitransaction_mode: bool
    threshold_encryption: bool

    def to_tuples(self) -> list:
        return [
            ('multitr', bool_to_bytes(self.multitransaction_mode)),
            ('encrypt', bool_to_bytes(self.threshold_encryption))
        ]


def parse_schain_options(raw_options: list) -> SchainOptions:
    """
    Parses raw sChain options from smart contracts (list of tuples).
    Returns default values if nothing is set on contracts.
    """
    if len(raw_options) == 0:
        return get_default_schain_options()
    return SchainOptions(
        multitransaction_mode=bytes_to_bool(raw_options[0][1]),
        threshold_encryption=bytes_to_bool(raw_options[1][1])
    )


def get_default_schain_options() -> SchainOptions:
    return SchainOptions(
        multitransaction_mode=False,
        threshold_encryption=False
    )


def bool_to_bytes(bool_value: bool) -> bytes:
    return bool_value.to_bytes(1, byteorder='big')


def bytes_to_bool(bytes_value: bytes) -> bool:
    return bool(int.from_bytes(bytes_value, 'big'))



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE base contract class """

import logging
from functools import wraps
from typing import Dict, Optional

from web3 import Web3

import skale.config as config
from skale.transactions.result import (
    TxRes,
    is_success,
    is_success_or_not_performed
)
from skale.transactions.tools import make_dry_run_call, transaction_from_method
from skale.utils.web3_utils import (
    DEFAULT_BLOCKS_TO_WAIT,
    get_eth_nonce,
    MAX_WAITING_TIME,
    wait_for_confirmation_blocks
)

from skale.utils.helper import to_camel_case


logger = logging.getLogger(__name__)


def execute_dry_run(skale, method, custom_gas_limit, value=0) -> tuple:
    dry_run_result = make_dry_run_call(skale, method, custom_gas_limit, value)
    estimated_gas_limit = None
    if is_success(dry_run_result):
        estimated_gas_limit = dry_run_result['payload']
    return dry_run_result, estimated_gas_limit


def transaction_method(transaction):
    @wraps(transaction)
    def wrapper(
        self,
        *args,
        wait_for=True,
        blocks_to_wait=DEFAULT_BLOCKS_TO_WAIT,
        timeout=MAX_WAITING_TIME,
        gas_limit=None,
        gas_price=None,
        nonce=None,
        max_fee_per_gas=None,
        max_priority_fee_per_gas=None,
        value=0,
        dry_run_only=False,
        skip_dry_run=False,
        raise_for_status=True,
        multiplier=None,
        priority=None,
        confirmation_blocks=0,
        meta: Optional[Dict] = None,
        **kwargs
    ):
        method = transaction(self, *args, **kwargs)
        dry_run_result, tx_hash, receipt = None, None, None

        nonce = get_eth_nonce(self.skale.web3, self.skale.wallet.address)

        # Make dry_run and estimate gas limit
        estimated_gas_limit = None
        if not skip_dry_run and not config.DISABLE_DRY_RUN:
            dry_run_result, estimated_gas_limit = execute_dry_run(
                self.skale, method, gas_limit, value
            )

        gas_limit = gas_limit or estimated_gas_limit or \
            config.DEFAULT_GAS_LIMIT

        gas_price = gas_price or config.DEFAULT_GAS_PRICE_WEI or \
            self.skale.gas_price
        # Send transaction
        should_send_transaction = not dry_run_only and \
            is_success_or_not_performed(dry_run_result)

        if should_send_transaction:
            tx = transaction_from_method(
                method=method,
                gas_limit=gas_limit,
                gas_price=gas_price,
                max_fee_per_gas=max_fee_per_gas,
                max_priority_fee_per_gas=max_priority_fee_per_gas,
                nonce=nonce,
                value=value
            )
            method_name = f'{self.name}.{method.abi.get("name")}'
            tx_hash = self.skale.wallet.sign_and_send(
                tx,
                multiplier=multiplier,
                priority=priority,
                method=method_name,
                meta=meta
            )
            if wait_for:
                receipt = self.skale.wallet.wait(tx_hash)
            if confirmation_blocks:
                wait_for_confirmation_blocks(
                    self.skale.web3,
                    confirmation_blocks
                )

        tx_res = TxRes(dry_run_result, tx_hash, receipt)

        if raise_for_status:
            tx_res.raise_for_status()
        return tx_res

    return wrapper


class BaseContract:
    def __init__(self, skale, name, address, abi):
        self.skale = skale
        self.name = name
        self.address = Web3.to_checksum_address(address)
        self.contract = skale.web3.eth.contract(address=self.address, abi=abi)

    def __getattr__(self, attr):
        """Fallback for contract calls"""
        logger.debug("Calling contract function: %s", attr)

        def wrapper(*args, **kw):
            logger.debug('called with %r and %r' % (args, kw))
            camel_case_fn_name = to_camel_case(attr)
            if hasattr(self.contract.functions, camel_case_fn_name):
                return getattr(self.contract.functions,
                               camel_case_fn_name)(*args, **kw).call()
            if hasattr(self.contract.functions, attr):
                return getattr(self.contract.functions,
                               attr)(*args, **kw).call()
            raise AttributeError(attr)
        return wrapper



import skale.contracts.manager.nodes as nodes
from skale.schain_config.generator import get_nodes_for_schain, get_schain_nodes_with_schains
from tests.constants import MIN_NODES_IN_SCHAIN


def test_get_nodes_for_schain(skale, schain):
    schain_name = schain
    schain_nodes = get_nodes_for_schain(skale, schain_name)
    fields_with_id = nodes.FIELDS.copy()
    fields_with_id.append('id')

    assert len(schain_nodes) >= MIN_NODES_IN_SCHAIN
    assert list(schain_nodes[0].keys()) == fields_with_id


def test_get_schain_nodes_with_schains(skale, schain):
    schain_name = schain
    nodes_with_schains = get_schain_nodes_with_schains(skale, schain_name)
    assert isinstance(nodes_with_schains[0]['schains'], list)
    assert isinstance(nodes_with_schains[0]['schains'][0]['mainnetOwner'], str)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE web3 utilities """

import logging
import os
import time
from typing import Iterable
from urllib.parse import urlparse

from eth_keys import keys
from web3 import Web3, WebsocketProvider, HTTPProvider
from web3.exceptions import TransactionNotFound
from web3.middleware import (
    attrdict_middleware,
    geth_poa_middleware,
    http_retry_request_middleware
)

import skale.config as config
from skale.transactions.exceptions import TransactionFailedError
from skale.utils.constants import GAS_PRICE_COEFFICIENT
from skale.utils.helper import is_test_env
from skale.transactions.exceptions import TransactionNotMinedError


logger = logging.getLogger(__name__)


WS_MAX_MESSAGE_DATA_BYTES = 5 * 1024 * 1024
MAX_WAITING_TIME = 3 * 60 * 60  # 3 hours
BLOCK_WAITING_TIMEOUT = 1
DEFAULT_HTTP_TIMEOUT = 120
DEFAULT_BLOCKS_TO_WAIT = 50


def get_provider(endpoint, timeout=DEFAULT_HTTP_TIMEOUT, request_kwargs={}):
    scheme = urlparse(endpoint).scheme
    if scheme == 'ws' or scheme == 'wss':
        kwargs = request_kwargs or {'max_size': WS_MAX_MESSAGE_DATA_BYTES}
        return WebsocketProvider(endpoint, websocket_timeout=timeout,
                                 websocket_kwargs=kwargs)

    if scheme == 'http' or scheme == 'https':
        kwargs = {'timeout': timeout, **request_kwargs}
        return HTTPProvider(endpoint, request_kwargs=kwargs)

    raise Exception(
        'Wrong endpoint option.'
        'Supported endpoint schemes: http/https/ws/wss'
    )


class EthClientOutdatedError(Exception):
    pass


def get_last_known_block_number(state_path: str) -> int:
    if not os.path.isfile(state_path):
        return 0
    with open(state_path) as last_block_file:
        return int(last_block_file.read())


def save_last_known_block_number(state_path: str, block_number: int) -> None:
    with open(state_path, 'w') as last_block_file:
        last_block_file.write(str(block_number))


def outdated_client_time_msg(method, current_time, latest_block_timestamp, allowed_ts_diff):
    return f'{method} failed; \
current_time: {current_time}, latest_block_timestamp: {latest_block_timestamp}, \
allowed_ts_diff: {allowed_ts_diff}'


def outdated_client_file_msg(method, latest_block_number, saved_number, state_path):
    return f'{method} failed: latest_block_number: {latest_block_number}, \
        saved_number: {saved_number}, state_path: {state_path}'


def make_client_checking_middleware(allowed_ts_diff: int,
                                    state_path: str = None):
    def eth_client_checking_middleware(make_request, web3):
        def middleware(method, params):
            if method in ('eth_block_number', 'eth_getBlockByNumber'):
                response = make_request(method, params)
            else:
                latest_block = web3.eth.get_block('latest')
                current_time = time.time()

                ts_diff = current_time - latest_block['timestamp']
                if not is_test_env():
                    ts_diff = abs(ts_diff)

                if ts_diff > allowed_ts_diff:
                    raise EthClientOutdatedError(outdated_client_time_msg(
                        method,
                        current_time,
                        latest_block['timestamp'],
                        allowed_ts_diff
                    ))

                if state_path:
                    saved_number = get_last_known_block_number(state_path)
                    if latest_block['number'] < saved_number:
                        raise EthClientOutdatedError(outdated_client_file_msg(
                            method,
                            latest_block['number'],
                            saved_number,
                            state_path
                        ))
                    save_last_known_block_number(state_path,
                                                 latest_block['number'])
                response = make_request(method, params)
            return response
        return middleware
    return eth_client_checking_middleware


def init_web3(endpoint: str,
              provider_timeout: int = DEFAULT_HTTP_TIMEOUT,
              middlewares: Iterable = None,
              state_path: str = None, ts_diff: int = None):
    if not middlewares:
        ts_diff = ts_diff or config.ALLOWED_TS_DIFF
        state_path = state_path or config.LAST_BLOCK_FILE
        if not ts_diff == config.NO_SYNC_TS_DIFF:
            sync_middleware = make_client_checking_middleware(ts_diff, state_path)
            middewares = (
                http_retry_request_middleware,
                sync_middleware,
                attrdict_middleware
            )
        else:
            middewares = (
                http_retry_request_middleware,
                attrdict_middleware
            )

    provider = get_provider(endpoint, timeout=provider_timeout)
    web3 = Web3(provider)
    # required for rinkeby
    web3.middleware_onion.inject(geth_poa_middleware, layer=0)
    for middleware in middewares:
        web3.middleware_onion.add(middleware)  # todo: may cause issues
    return web3


def get_receipt(web3, tx):
    return web3.eth.get_transaction_receipt(tx)


def get_eth_nonce(web3, address):
    return web3.eth.get_transaction_count(address)


def wait_for_receipt_by_blocks(
    web3,
    tx,
    blocks_to_wait=DEFAULT_BLOCKS_TO_WAIT,
    timeout=MAX_WAITING_TIME
):
    blocks_to_wait = blocks_to_wait or DEFAULT_BLOCKS_TO_WAIT
    timeout = timeout or MAX_WAITING_TIME
    previous_block = web3.eth.block_number
    current_block = previous_block
    wait_start_time = time.time()
    while time.time() - wait_start_time < timeout and \
            current_block <= previous_block + blocks_to_wait:
        try:
            receipt = get_receipt(web3, tx)
        except TransactionNotFound:
            receipt = None
        if receipt is not None:
            return receipt
        current_block = web3.eth.block_number
        time.sleep(3)
    raise TransactionNotMinedError(
        f'Transaction with hash: {tx} not found in {blocks_to_wait} blocks.'
    )


def wait_receipt(web3, tx, retries=30, timeout=5):
    for _ in range(0, retries):
        try:
            receipt = get_receipt(web3, tx)
        except TransactionNotFound:
            receipt = None
        if receipt is not None:
            return receipt
        time.sleep(timeout)  # pragma: no cover
    raise TransactionNotMinedError(
        f'Transaction with hash: {tx} not mined after {retries} retries.'
    )


def check_receipt(receipt, raise_error=True):
    if receipt['status'] != 1:  # pragma: no cover
        if raise_error:
            raise TransactionFailedError(
                f'Transaction failed, see receipt {receipt}'
            )
        else:
            return False
    return True


def wait_for_confirmation_blocks(
    web3,
    blocks_to_wait,
    timeout=MAX_WAITING_TIME,
    request_timeout=5
):
    current_block = start_block = web3.eth.block_number
    logger.info(
        f'Current block number is {current_block}, '
        f'waiting for {blocks_to_wait} confimration blocks to be mined'
    )
    wait_start_time = time.time()
    while time.time() - wait_start_time < timeout and \
            current_block <= start_block + blocks_to_wait:
        current_block = web3.eth.block_number
        time.sleep(request_timeout)


def private_key_to_public(pr):
    pr_bytes = Web3.to_bytes(hexstr=pr)
    pk = keys.PrivateKey(pr_bytes)
    return pk.public_key


def public_key_to_address(pk):
    hash = Web3.keccak(hexstr=str(pk))
    return Web3.to_hex(hash[-20:])


def private_key_to_address(pr):
    pk = private_key_to_public(pr)
    return public_key_to_address(pk)


def to_checksum_address(address):
    return Web3.to_checksum_address(address)


def wallet_to_public_key(wallet):
    if isinstance(wallet, dict):
        return private_key_to_public(wallet['private_key'])
    else:
        return wallet['public_key']


def default_gas_price(web3: Web3) -> int:
    return web3.eth.gas_price * GAS_PRICE_COEFFICIENT



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" Nodes.sol functions """

import socket
from enum import IntEnum

from Crypto.Hash import keccak
from web3.exceptions import BadFunctionCallOutput, ContractLogicError

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes

from skale.utils.exceptions import InvalidNodeIdError
from skale.utils.helper import format_fields

FIELDS = [
    'name', 'ip', 'publicIP', 'port', 'start_block',
    'last_reward_date', 'finish_time', 'status', 'validator_id', 'publicKey', 'domain_name'
]


class NodeStatus(IntEnum):
    ACTIVE = 0
    LEAVING = 1
    LEFT = 2
    IN_MAINTENANCE = 3


class Nodes(BaseContract):
    def __get_raw(self, node_id):
        try:
            return self.contract.functions.nodes(node_id).call()
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def __get_raw_w_pk(self, node_id):
        raw_node_struct = self.__get_raw(node_id)
        raw_node_struct.append(self.get_node_public_key(node_id))
        return raw_node_struct

    def __get_raw_w_pk_w_domain(self, node_id):
        raw_node_struct_w_pk = self.__get_raw_w_pk(node_id)
        raw_node_struct_w_pk.append(self.get_domain_name(node_id))
        return raw_node_struct_w_pk

    @format_fields(FIELDS)
    def get(self, node_id):
        return self.__get_raw_w_pk_w_domain(node_id)

    @format_fields(FIELDS)
    def get_by_name(self, name):
        name_hash = self.name_to_id(name)
        _id = self.contract.functions.nodesNameToIndex(name_hash).call()
        return self.__get_raw_w_pk_w_domain(_id)

    def get_nodes_number(self):
        return self.contract.functions.getNumberOfNodes().call()

    def get_active_node_ids(self):
        nodes_number = self.get_nodes_number()
        return [
            node_id
            for node_id in range(0, nodes_number)
            if self.get_node_status(node_id) == NodeStatus.ACTIVE
        ]

    def get_active_node_ips(self):
        nodes_number = self.get_nodes_number()
        return [
            self.get(node_id)['ip']
            for node_id in range(0, nodes_number)
            if self.get_node_status(node_id) == NodeStatus.ACTIVE
        ]

    def name_to_id(self, name):
        keccak_hash = keccak.new(data=name.encode("utf8"), digest_bits=256)
        return keccak_hash.digest()

    def is_node_name_available(self, name):
        node_id = self.name_to_id(name)
        return not self.contract.functions.nodesNameCheck(node_id).call()

    def is_node_ip_available(self, ip):
        ip_bytes = socket.inet_aton(ip)
        return not self.contract.functions.nodesIPCheck(ip_bytes).call()

    def node_name_to_index(self, name):
        name_hash = self.name_to_id(name)
        return self.contract.functions.nodesNameToIndex(name_hash).call()

    def get_node_status(self, node_id):
        try:
            return self.contract.functions.getNodeStatus(node_id).call()
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def get_node_finish_time(self, node_id):
        try:
            return self.contract.functions.getNodeFinishTime(node_id).call()
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def __get_node_public_key_raw(self, node_id):
        try:
            return self.contract.functions.getNodePublicKey(node_id).call()
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def get_node_public_key(self, node_id):
        raw_key = self.__get_node_public_key_raw(node_id)
        key_bytes = raw_key[0] + raw_key[1]
        return self.skale.web3.to_hex(key_bytes)

    def get_validator_node_indices(self, validator_id: int) -> list:
        """Returns list of node indices to the validator

        :returns: List of trusted node indices
        :rtype: list
        """
        return self.contract.functions.getValidatorNodeIndexes(validator_id).call()

    @transaction_method
    def set_node_in_maintenance(self, node_id):
        return self.contract.functions.setNodeInMaintenance(node_id)

    @transaction_method
    def remove_node_from_in_maintenance(self, node_id):
        return self.contract.functions.removeNodeFromInMaintenance(node_id)

    @transaction_method
    def set_domain_name(self, node_id: int, domain_name: str):
        return self.contract.functions.setDomainName(node_id, domain_name)

    def get_domain_name(self, node_id: int):
        return self.contract.functions.getNodeDomainName(node_id).call()

    @transaction_method
    def grant_role(self, role: bytes, owner: str) -> TxRes:
        return self.contract.functions.grantRole(role, owner)

    def has_role(self, role: bytes, address: str) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    def node_manager_role(self):
        return self.contract.functions.NODE_MANAGER_ROLE().call()

    def compliance_role(self):
        return self.contract.functions.COMPLIANCE_ROLE().call()

    @transaction_method
    def init_exit(self, node_id: int) -> TxRes:
        return self.contract.functions.initExit(node_id)

    @transaction_method
    def change_ip(self, node_id: int, ip: bytes, public_ip: bytes) -> TxRes:
        return self.contract.functions.changeIP(node_id, ip, public_ip)
